/* unicap
 *
 * Copyright (C) 2004 Arne Caspari ( arne_caspari@users.sourceforge.net )
 *
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <linux/types.h>
#include <stdlib.h>
#include <linux/types.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/time.h>
#include <time.h>

#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#include <X11/extensions/Xvlib.h>
#include <X11/X.h>

#include <sys/ipc.h>
#include <sys/shm.h>

#include "colorspace.h"

#include "unicap.h"
#include "ucil.h"
#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <semaphore.h>


#include <unicapgtk.h>
#include "backend.h"
#include "debug.h"

#define MAX_IMAGE_BUFFERS 4
#define MAX_XV_PORTS 64

static int g_used_ports[MAX_XV_PORTS];
static int g_initialized = 0;

typedef struct _xv_fourcc
{
      unsigned int fourcc;
      char guid[16];
} xv_fourcc_t;

struct xv_backend_data
{
      XvAdaptorInfo *p_adaptor_info;
      unsigned int num_adaptors;
      XvPortID port_id;
		
      unsigned int xv_mode_id;	
		
      XvImage *image[MAX_IMAGE_BUFFERS];
      int use_shm;
      XShmSegmentInfo shminfo[MAX_IMAGE_BUFFERS];
      struct timeval fill_times[MAX_IMAGE_BUFFERS];
      unicap_format_t image_formats[MAX_IMAGE_BUFFERS];
		
      int width;
      int height;
      
      gint output_width;
      gint output_height;
      gint crop_x;
      gint crop_y;
      gint crop_w;
      gint crop_h;

      int window_width;
      int window_height;
		
      Atom atom_colorkey;
      Atom atom_brightness;
      Atom atom_hue;
      Atom atom_contrast;
		
      int brightness_min;
      int brightness_max;
      int hue_min;
      int hue_max;
      int contrast_max;
      int contrast_min;

      unsigned int colorkey;

      volatile int current_buffer;
		
      Display *display;
      Drawable drawable;
      GC       gc;
      GdkGC   *gdkgc;

      GtkWidget *widget;
      GdkWindow *overlay_window;
      GC overlay_gc;

      gboolean scale_to_fit;

      gint redisplay_timerid;
      gboolean pause_state;

      sem_t sema;

};


static XvImage *allocate_image( Display *display, 
				XvPortID port_id, 
				unsigned int xv_mode_id,
				XShmSegmentInfo *shminfo, 
				int width, 
				int height, 
				int *use_shm )
{
  XvImage *image = 0;
  if( shminfo )
  {
     image = XvShmCreateImage( display, 
			       port_id, 
			       xv_mode_id, 
			       (char*)NULL, 
			       width,
			       height, 
			       shminfo );
	
     if( image )
     {
	shminfo->shmid = shmget( IPC_PRIVATE, image->data_size, IPC_CREAT | 0777 );
	if( shminfo->shmid == -1 )
	{
	   return 0;
	}
	shminfo->shmaddr = image->data = shmat( shminfo->shmid, 0, 0 );
	shmctl(shminfo->shmid, IPC_RMID, 0);

	if( shminfo->shmaddr == ( void * ) -1 )
	{
	   return 0;
	}
	
	shminfo->readOnly = False;
	if( !XShmAttach( display, shminfo ) )
	{
	   shmdt( shminfo->shmaddr );
	   XFree( image );
	   return 0;
	}
	
	*use_shm = 1;
     }
  }
  else
  {
     image = XvCreateImage( display, 
			    port_id, 
			    xv_mode_id, 
			    (char*)NULL, 
			    width,
			    height );
     if( image )
     {
	g_assert( image->data_size );
	image->data = g_slice_alloc( image->data_size );
     }

     *use_shm = 0;
  }
  
  return image;
}


gint backend_get_merit()
{
   return BACKEND_MERIT_XV;
}


static gboolean redraw_timeout( gpointer _data )
{
   backend_redraw( _data );
   return TRUE;
}

static GdkGC *create_gc( GtkWidget *widget, int r, int g, int b )
{
   GdkGCValues values;
   GdkColormap *colormap;
   GdkGC *gc;

   values.foreground.red = ( r << 8 ) | r;
   values.foreground.green = ( g << 8 ) | g ;
   values.foreground.blue = ( b << 8 ) | b;
   values.foreground.pixel = 0;

   colormap = gtk_widget_get_colormap( widget );
   gdk_rgb_find_color( colormap, &values.foreground );

   gc = gdk_gc_new_with_values( widget->window,
				&values, GDK_GC_FOREGROUND );

   return gc;
}


static void set_colorkey( struct xv_backend_data *data,  unsigned int r, unsigned int g, unsigned int b )
{
  unsigned int val = ( ((r & 0xff) << 16) | ((g & 0xff) << 8) | (b & 0xff) );

  if( data->atom_colorkey )
  {
     XvSetPortAttribute( data->display,
                         data->port_id,
                         data->atom_colorkey,
                         val );
     XvGetPortAttribute( data->display,
                         data->port_id, 
                         data->atom_colorkey,
                         (int*)&data->colorkey );
  }
  else
  {
     data->colorkey = val;
  }
}


void backend_size_allocate( gpointer _data, GtkWidget *widget, GtkAllocation *allocation )
{
   struct xv_backend_data *data = _data;

   if( data->scale_to_fit )
   {      
      data->output_width = allocation->width;
      data->output_height = allocation->height;
   }
}

gboolean backend_init( GtkWidget *widget, unicap_format_t *format, gpointer *_data, GError **err )
{
   unsigned int version, release;
   unsigned int request_base, event_base, error_base;
	
   int i;
   Display *dpy;
   struct xv_backend_data *data;
   int use_shm = 1;
   unsigned int target_fourcc = format->fourcc;

   int numPorts;

   GKeyFile *keyfile;
   gchar *keyfile_path;
   gchar *userconf_path;
   gchar *cval;

   unsigned int r,g,b;
   XGCValues values;

   TRACE( "Backend xv: init\n" );

   if( !g_initialized )
   {
      TRACE( "Initialize Xv port list\n" );
      memset( g_used_ports, 0x0, sizeof( g_used_ports ) );
      g_initialized = 1;
   }

   g_assert( GTK_WIDGET_REALIZED( widget ) );   

   data = g_new0( struct xv_backend_data, 1 );
   dpy = GDK_DISPLAY_XDISPLAY(gtk_widget_get_display( widget ) );
   if( XvQueryExtension( dpy, 
			 &version, &release, &request_base, &event_base, &error_base ) != Success )
   {
      TRACE( "XvQueryExtension failed\n" );
      free( data );
      return FALSE;
   }
   if( XvQueryAdaptors( dpy, DefaultRootWindow( dpy ), &data->num_adaptors, &data->p_adaptor_info ) != Success )
   {
      TRACE( "XvQueryAdaptors failed\n" );
      free( data );
      return FALSE;
   }
   if( data->num_adaptors == 0 )
   {
      XFree( data->p_adaptor_info );
      free( data );
      return FALSE;
   }
	

   data->port_id = -1;

   userconf_path = unicapgtk_get_user_config_path( );
   keyfile_path = g_build_path( G_DIR_SEPARATOR_S, userconf_path, "unicapgtk.conf", NULL );
   g_free( userconf_path );
/*    if( !g_file_test( keyfile_path, G_FILE_TEST_EXISTS ) ) */
/*    { */
/*       g_free( keyfile_path ); */
/*       keyfile_path = g_build_path( G_DIR_SEPARATOR_S, SYSCONFDIR, "unicap", "unicapgtk.conf", NULL ); */
/*    } */
   
   keyfile = g_key_file_new();
   g_key_file_load_from_file( keyfile, keyfile_path, 0, NULL );
   g_free( keyfile_path );

   cval = g_key_file_get_value( keyfile, "display_backend_xv", "force_fourcc", NULL );
   if( cval )
   {
      target_fourcc = cval[0] + (cval[1]<<8) + (cval[2]<<16) + (cval[3]<<24);
      

      if( !ucil_conversion_supported( target_fourcc, format->fourcc ) )
      {
	 g_key_file_free( keyfile );
	 TRACE( "No conversion from format->fourcc to keyfile fourcc\n" );
	 XFree( data->p_adaptor_info );
	 free( data );
	 return FALSE;
      }
   }
   
   numPorts = 0;
   for( i = 0; i < data->num_adaptors; i++ )
   {
      numPorts += data->p_adaptor_info[i].num_ports;
   }
   

   for( i = 0; ( i < numPorts ) && ( data->port_id == -1 ) ; i++ )
   {

      XvAttribute *at;
      int num_attributes;
      int attribute;

      XvImageFormatValues *xvimage_formats;
      int num_xvimage_formats;
      int f;
      XvPortID tmp_port_id = -1;

      tmp_port_id = data->p_adaptor_info[0].base_id + i;
      TRACE( "port id: %08x\n:", (unsigned int )tmp_port_id );
      

      xvimage_formats = XvListImageFormats( dpy, 
					    tmp_port_id, 
					    &num_xvimage_formats );
      for( f = 0; f < num_xvimage_formats; f++ )
      {
	 char imageName[5] = {0, 0, 0, 0, 0};
	 memcpy(imageName, &(xvimage_formats[f].id), 4);
	 TRACE( "      id: 0x%x", xvimage_formats[f].id);
	 if( isprint( imageName[0]) && isprint(imageName[1]) &&
	     isprint( imageName[2]) && isprint(imageName[3])) 
	 {
	    DBGOUT(" (%s)\n", imageName);
	 } else {
	    DBGOUT("\n");
	 }
	 if( xvimage_formats[f].id == target_fourcc )
	 {
	    data->xv_mode_id = target_fourcc;
	    data->port_id = tmp_port_id;
	    TRACE( " using fourcc: %08x\n", target_fourcc );
	    break;
	 }
      }

      XFree( xvimage_formats );

      if( data->xv_mode_id != target_fourcc )
      {
	 continue;
      }


      if( !g_used_ports[i] )
      {
	 if( XvGrabPort( dpy, tmp_port_id, CurrentTime ) 
	     != Success )
	 {
	    TRACE( "can not grab port: %d\n", (unsigned int)tmp_port_id );
	    data->port_id = -1;
	    continue;
	 }
	 g_used_ports[i] = 1;
      }
      else
      {
	 TRACE( "port %d is in use!\n", i );
	 data->port_id = -1;
	 continue;
      }
      
      at = XvQueryPortAttributes( dpy, tmp_port_id, &num_attributes );
/*       TRACE( "num_attributes: %d\n", num_attributes ); */
      for( attribute = 0; attribute < num_attributes; attribute++ )
      {
	 
	 if( !strcmp( at[attribute].name, "XV_COLORKEY" ) )
	 {
	    int val;
	    Atom atom;
	    atom = (Atom)XInternAtom( dpy, at[attribute].name, 0 );
	    XvGetPortAttribute( dpy, 
				tmp_port_id, 
				atom, 
				&val );
	    data->colorkey = val;
	    data->atom_colorkey = atom;
	 }
	 
	 if( !strcmp( at[attribute].name, "XV_DOUBLE_BUFFER" ) )
	 {
	    Atom _atom;
	    _atom = XInternAtom( dpy, at[attribute].name, 0 );
	    XvSetPortAttribute( dpy, tmp_port_id, _atom, 1 );
	    TRACE( "Xv: DOUBLE_BUFFER available\n" );
	 }

	 cval = g_key_file_get_value( keyfile, "display_backend_xv", at[attribute].name, NULL );
	 if( cval )
	 {
	    int val;
	    Atom atom;
	    atom = (Atom)XInternAtom( dpy, at[attribute].name, 0 );
	    val = atoi( cval );
	    XvSetPortAttribute( dpy, tmp_port_id, atom, val );
	    TRACE( "Set Attribute for: %s to %d\n", at[attribute].name, val );
	 }
      }
      XFree( at );
   }

   if( data->port_id == -1 )
   {
      TRACE( "No suitable port found\n" );
      XFree( data->p_adaptor_info );
      free( data );
      return FALSE;
   }

   TRACE( "Using port: %d\n", (unsigned int )data->port_id );

   data->display = dpy;
   data->width = format->size.width;
   data->height = format->size.height;
   data->output_width = data->width;
   data->output_height = data->height;
   data->crop_x = 0;
   data->crop_y = 0;
   data->crop_w = data->width;
   data->crop_h = data->height;
   data->widget = widget;

   data->current_buffer = -1;
   
   set_colorkey( data, 0x5, 0x5, 0xfe );

   if( data->atom_colorkey )
   {
      r = ( data->colorkey >> 16 ) & 0xff;
      g = ( data->colorkey >>  8 ) & 0xff;
      b = ( data->colorkey ) & 0xff;
      data->gdkgc = create_gc( widget, r, g, b );
   }

   for( i = 0; i < MAX_IMAGE_BUFFERS; i++ )
   {
      memset( &data->shminfo[i], 0, sizeof( XShmSegmentInfo ) );
   
      data->image[i] = allocate_image( data->display, 
				       data->port_id,
				       data->xv_mode_id, 
				       &data->shminfo[i], 
				       data->width, 
				       data->height, 
				       &use_shm );
      if( !data->image[i] )
      {
	 g_assert( "Failed to allocate image" == FALSE );
      }
   }

   data->use_shm = use_shm;

   *_data = data;

   XSync( dpy, FALSE );

   g_key_file_free( keyfile );

   sem_init( &data->sema, 0, 1 );

   data->overlay_window = widget->window;

   data->overlay_gc = XCreateGC( data->display,
				 GDK_WINDOW_XWINDOW( data->overlay_window ),
				 0, &values );
   g_assert( data->overlay_gc );

   gdk_window_show( data->overlay_window );

   return TRUE;
}

void backend_destroy( gpointer _data )
{
   struct xv_backend_data *data = _data;
   
   int i;

   XSync( data->display, FALSE );

   XFreeGC( data->display, data->overlay_gc );
   if( data->gdkgc )
   {
      g_object_unref( data->gdkgc );
   }
 
   for( i = 0; i < MAX_IMAGE_BUFFERS; i++ )
   {
      shmdt( data->shminfo[i].shmaddr );
      XFree( data->image[i] );
   }

   XvUngrabPort( data->display, data->port_id, CurrentTime );

   XFree( data->p_adaptor_info );

   g_free( data );
}


void backend_set_crop( gpointer _data, int crop_x, int crop_y, int crop_w, int crop_h )
{
   struct xv_backend_data *data = _data;
   
   data->crop_x = crop_x;
   data->crop_y = crop_y;
   data->crop_w = crop_w;
   data->crop_h = crop_h;
}

void backend_redraw( gpointer _data )
{
   struct xv_backend_data *data = _data;

/*    printf( "redraw %d\n", data->current_buffer ); */

   if( data->atom_colorkey )
   {
      int width, height;
      
      width = data->widget->allocation.width < data->output_width ? data->widget->allocation.width : data->output_width;
      height = data->widget->allocation.height < data->output_height ? data->widget->allocation.height : data->output_height;
      
      gdk_draw_rectangle( data->widget->window,
			  data->gdkgc,
			  TRUE,
			  0, 0,
			  width, height );
      
   }

   if( data->current_buffer >= 0 )
   {
      if( data->use_shm )
      {
/* 	 printf( "shm %dx%d\n", data->crop_x, data->crop_y ); */
	 XvShmPutImage( data->display,
			data->port_id,
			GDK_WINDOW_XWINDOW( data->overlay_window ),
			data->overlay_gc,
			data->image[data->current_buffer],
			data->crop_x, data->crop_y, data->crop_w, data->crop_h,
			0, 0, data->output_width, data->output_height,
			1 );
      }
      else
      {
/* 	 printf( "!shm\n" ); */
	 
	 XvPutImage( data->display,
		     data->port_id,
		     GDK_WINDOW_XWINDOW( data->overlay_window ),
		     data->overlay_gc,
		     data->image[data->current_buffer],
		     data->crop_x, data->crop_y, data->crop_w, data->crop_h,
		     0, 0, data->output_width, data->output_height );
      }
   }
   

/*    printf( "-redraw\n" ); */
}

void backend_update_image( gpointer _data, unicap_data_buffer_t *data_buffer, GError **err )
{
   struct xv_backend_data *data = _data;
   int tmp;

/*    printf( "ui\n" ); */
   

   tmp = ( data->current_buffer + 1 ) % MAX_IMAGE_BUFFERS;

   if( data_buffer->format.fourcc != data->xv_mode_id )
   {
      unicap_data_buffer_t tmp_buffer;
      unicap_copy_format( &tmp_buffer.format, &data_buffer->format );
      tmp_buffer.format.bpp = 24;
      tmp_buffer.format.fourcc = data->xv_mode_id;
      tmp_buffer.data = ( unsigned char * )data->image[tmp]->data;
      
      ucil_convert_buffer( &tmp_buffer, data_buffer );

   }
   else
   {
      g_memmove( data->image[tmp]->data, 
		 data_buffer->data, 
		 data_buffer->buffer_size );
   }
   
   unicap_copy_format( &data->image_formats[tmp], &data_buffer->format );
   memcpy( &data->fill_times[tmp], &data_buffer->fill_time, sizeof( struct timeval) );

}

void backend_display_image( gpointer _data )
{
   struct xv_backend_data *data = _data;

   data->current_buffer = ( data->current_buffer + 1 ) % MAX_IMAGE_BUFFERS;
   backend_redraw( _data );
}

void backend_expose_event( gpointer _data, GtkWidget *da, GdkEventExpose *event )
{
   struct xv_backend_data *data = _data;

   if( data->gdkgc )
   {
      GdkGCValues values;
      
      gdk_gc_get_values( data->gdkgc, &values );      

      gdk_draw_rectangle( da->window,
			  data->gdkgc,
			  TRUE,
			  event->area.x, event->area.y,
			  event->area.width, event->area.height );
   }
   

   backend_redraw( _data );
}

void backend_get_image_data( gpointer _data, unicap_data_buffer_t *data_buffer, int b )
{
   struct xv_backend_data *data = _data;
   int tmp;
   
   tmp = ( data->current_buffer + b ) % MAX_IMAGE_BUFFERS;

/*    unicap_void_format( &data_buffer->format ); */
/*    data_buffer->format.fourcc = data->xv_mode_id; */
/*    data_buffer->format.bpp = 16; */
/*    data_buffer->format.size.width = data->width; */
/*    data_buffer->format.size.height = data->height; */
/*    data_buffer->format.buffer_size = data->width * data->height * 2; */
   data_buffer->data = (unsigned char*)data->image[tmp]->data;
   unicap_copy_format( &data_buffer->format, &data->image_formats[tmp] );
   data_buffer->buffer_size = data_buffer->format.buffer_size;
   memcpy( &data_buffer->fill_time, &data->fill_times[tmp], sizeof( struct timeval ) );
}

void backend_lock( gpointer _data )
{
   struct xv_backend_data *data = _data;
   sem_wait( &data->sema );
}

void backend_unlock( gpointer _data )
{
   struct xv_backend_data *data = _data;
   sem_post( &data->sema );
}

void backend_set_scale_to_fit( gpointer _data, gboolean scale_to_fit )
{
   struct xv_backend_data *data = _data;

   data->scale_to_fit = scale_to_fit;
}


void backend_set_pause_state( gpointer _data, gboolean state )
{
   struct xv_backend_data *data = _data;

   if( data->redisplay_timerid )
   {
      g_source_remove( data->redisplay_timerid );
      data->redisplay_timerid = 0;
   }

   TRACE( "set pause: %d\n", state );
   
   if( state )
   {
      data->redisplay_timerid = g_timeout_add( 300, (GSourceFunc)redraw_timeout, _data );
   }

   data->pause_state = state;
   
}

guint backend_get_flags( gpointer _data )
{
   return BACKEND_FLAGS_SCALING_SUPPORTED;
}

